implementation module Geometry

import StdEnvExt

/**	scale_vector k {dx,dy}
		returns {k*dx,k*dy}
	scale_vector3D k {dxy,dz}
		returns {scale_vector k dxy,k*dz}
*/
scale_vector :: !Real !RVector -> RVector
scale_vector k {dx,dy}
	= {dx=k*dx, dy=k*dy}

scale_vector3D :: !Real !RVector3D -> RVector3D
scale_vector3D k {dxy,dz}
	= {dxy=scale_vector k dxy,dz=k*dz}
	
instance zero RVector   where	zero = {dx=zero,dy=zero}
instance zero RVector3D where	zero = {dxy=zero,dz=zero}
instance ~    RVector   where	~ v  = zero - v
instance ~    RVector3D where	~ v  = zero - v


// how far away from backwards (pi)?
// pi/1,5 = pi/0,5
angleHowFarFromPi :: !Angle -> Angle
angleHowFarFromPi a		= toRadian (if (a` > 180) (360 - a`) a`)
where
	a`					= fromRadian (abs a)

angleHowFarFromAngle :: !Angle !Angle -> Angle
angleHowFarFromAngle a b
	# a` = fromRadian (abs a)
	# b` = fromRadian (abs b)
	| a` > b`
		| a` - b` > 180	= toRadian (b` - a` + 360)
		| otherwise		= toRadian (a` - b`)
	| otherwise
		| b` - a` > 180	= toRadian (a` - b` + 360)
		| otherwise		= toRadian (b` - a`)

fromRadian :: !Radian -> Degrees
fromRadian a		= (toInt (a * (360.0 / (2.0*pi)))) mod 360

toRadian :: !Degrees -> Radian
toRadian a			= (toReal a) * pi / 180.0

/** move_point v p
		moves point p over vector v.
*/
move_point :: !RVector !Position -> Position
move_point {dx,dy} {px,py}
	= {px=px+dx,py=py+dy}

move_point3D :: !RVector3D !Position3D -> Position3D
move_point3D {dxy,dz} {pxy,pz}
	= {pxy=move_point dxy pxy,pz=pz+dz}
	
/**	between_points (a,b) c
		returns True iff c is on the line between a and b.
*/
between_points :: !(!Position,!Position) !Position -> Bool
between_points (a,b) c
	= point_in_rectangle (a,b) c && dcx / dcy == dx / dy
where
	[min_x,max_x:_]	= sort [a.px,b.px]
	[min_y,max_y:_]	= sort [a.py,b.py]
	(dx, dy)		= (a.px - b.px, a.py - b.py)
	(dcx,dcy)		= (a.px - c.px, a.py - c.py)


/** point_in_rectangle (a,b) c
		returns True if point c is inside the rectangle determined by 
		the diagonal corner points a and b.
*/
point_in_rectangle :: !(!Position,!Position) !Position -> Bool
point_in_rectangle (a,b) c
	= isbetween c.px min_x max_x && isbetween c.py min_y max_y
where
	(min_x,max_x)		= minmax (a.px,b.px)
	(min_y,max_y)		= minmax (a.py,b.py)

/**	point_to_rectangle (a,b) c
		returns c if (point_in_rectangle (a,b) c) and the projected point c` of c that is exactly
		on the closest edge of rectangle (a,b).
*/
point_to_rectangle :: !(!Position,!Position) !Position -> Position
point_to_rectangle (a,b) c
	| point_in_rectangle (a,b) c
						= c
	| otherwise			= toPosition c`
where
	(min_x,max_x)		= minmax (a.px,b.px)
	(min_y,max_y)		= minmax (a.py,b.py)
	left				= c.px <= min_x
	right				= c.px >= max_x
	above				= c.py <= min_y
	below				= c.py >= max_y
	
	c` | left  && above	= (min_x,min_y)
	   | right && above	= (max_x,min_y)
	   | left  && below	= (min_x,max_y)
	   | right && below = (max_x,max_y)
	   | above			= (c.px, min_y)
	   | below			= (c.px, max_y)
	   | left			= (min_x,c.py )
	   | right			= (max_x,c.py )
	   | otherwise		= abort ("unsuspected error; please rotate with angles between pi and -pi\n")

	
instance toRVector Real where	// interpret Real as angle in radians
	toRVector angle	= {dx=cos angle,dy=sin angle}


instance toRVector Position where
	toRVector p = {dx=p.px,dy=p.py}
	
inRadiusOfPosition :: !Position3D !XRadius YRadius ZRadius !Position -> Bool
inRadiusOfPosition pos3d=:{pxy=pxy=:{px,py},pz} xr yr zr pos
	= pos.px+xr >= px && px >= pos.px-xr && pos.py+yr >= py && py >= pos.py-yr && pz <= zr
		
inCircleRadiusOfPosition :: !Position3D !XRadius ZRadius !Position -> Bool
inCircleRadiusOfPosition {pxy,pz} r zr pos
	= dist pxy pos <= r && pz <= zr
	
	
/** size_vector {dx,dy} = sqrt (dx^2 + dy^2)
	size_vector3D {dxy,dz} = sqrt (dxy.dx^2 + dxy.dy^2 + dz^2)
*/
size_vector :: !RVector -> Real
size_vector {dx,dy} = sqrt (dx^2.0 + dy^2.0)

size_vector3D :: !RVector3D -> Real
size_vector3D {dxy,dz} = sqrt (dxy.dx^2.0 + dxy.dy^2.0 + dz^2.0)

class dist a b :: !a !b -> Real
instance dist Real Real where
	dist a b = abs (a-b)
instance dist Position Position where
	dist a b = sqrt ((abs (a.px-b.px))^2.0 + (abs (a.py-b.py))^2.0)
instance dist Position3D Position3D where
	dist a b = sqrt ((abs (a.pxy.px-b.pxy.px))^2.0 + (abs (a.pxy.py-b.pxy.py))^2.0 + (abs (a.pz-b.pz))^2.0)
instance dist Position Position3D where
	dist a b = dist {pxy=a,pz=zero} b
instance dist Position3D Position where
	dist a b = dist a {pxy=b,pz=zero}


/** orthogonal a
		returns the left- and right- orthogonal angles to a
*/
orthogonal :: !Angle -> (!Angle,!Angle)
orthogonal a = (a+pi/4.0, a-pi/4.0)

instance zero Position where
	zero = {px=zero,py=zero}
instance zero Position3D where
	zero = {pxy=zero,pz=zero}
	
instance + RVector where
	(+) v1 v2 = {dx=v1.dx+v2.dx, dy=v1.dy+v2.dy}
instance + RVector3D where
	(+) v1 v2 = {dxy=v1.dxy+v2.dxy, dz=v1.dz+v2.dz}

instance - RVector where
	(-) v1 v2 = {dx=v1.dx-v2.dx, dy=v1.dy-v2.dy}
instance - RVector3D where
	(-) v1 v2 = {dxy=v1.dxy-v2.dxy, dz=v1.dz-v2.dz}
	
instance toPosition     (!Real,!Real)       where toPosition     (x,y)   = {px=x,py=y}
instance toPosition     Position3D          where toPosition     p3D     = p3D.pxy
instance fromPosition   (!Real,!Real)       where fromPosition   p2D     = (p2D.px,p2D.py)
instance fromPosition   Position3D          where fromPosition   p2D     = {zero & pxy=p2D}
instance toPosition3D   (!Real,!Real,!Real) where toPosition3D   (x,y,z) = {pxy=toPosition (x,y),pz=z}
instance toPosition3D   Position            where toPosition3D   p2D     = {zero & pxy=p2D}
instance fromPosition3D (!Real,!Real,!Real) where fromPosition3D p3D     = (p3D.pxy.px,p3D.pxy.py,p3D.pz)
instance fromPosition3D Position            where fromPosition3D p3D     = p3D.pxy

instance toString Speed      where toString {direction,velocity}= "{direction=" +++ toString direction +++ ",velocity=" +++ toString velocity +++ "}"
instance toString Speed3D    where toString {vxy,vz}			= "{vxy="       +++ toString vxy       +++ ",vz="       +++ toString vz       +++ "}"
instance toString Position   where toString {px, py}			= "{px="        +++ toString px        +++ ",py="       +++ toString py       +++ "}"
instance toString Position3D where toString {pxy,pz}			= "{pxy="       +++ toString pxy       +++ ",pz="       +++ toString pz       +++ "}"

instance zero Speed          where zero							= {direction=zero,velocity=zero}
instance zero Speed3D        where zero							= {vxy=zero,vz=zero}

instance == Speed            where (==) sp1 sp2					= sp1.direction == sp2.direction && sp1.velocity == sp2.velocity
instance == Speed3D          where (==) sp1 sp2					= sp1.vxy == sp2.vxy && sp1.vz == sp2.vz
instance == Position         where (==) p1  p2  				= p1.px   == p2.px   && p1.py  == p2.py
instance == Position3D       where (==) p1  p2  				= p1.pxy  == p2.pxy  && p1.pz  == p2.pz

instance toSpeed     Speed3D where toSpeed     s				= s.vxy
instance fromSpeed   Speed3D where fromSpeed   s				= {zero & vxy=s}
instance toSpeed3D   Speed   where toSpeed3D   s				= {zero & vxy=s}
instance fromSpeed3D Speed   where fromSpeed3D s				= s.vxy


oppositeAngle :: !Angle -> Angle
oppositeAngle a	
	| new_angle < ~pi	= new_angle + 2.0*pi
	| otherwise			= new_angle
where
	new_angle			= a - pi

angleWithObject :: !Position !Position -> Angle
angleWithObject base target
	| a >= (0.5*pi)	&& b >= zero	= ((~pi)+a)	// linksvoor,    naar boven,   negatieve hoek, -0.5*pi < hoek < zero
	| a <= (0.5*pi) && b >= zero	= (~pi)+b	// linksachter,  naar beneden, negatieve hoek, -0.5*pi < hoek < -pi
	| a <= (0.5*pi) && b <= zero	= pi+b		// rechtsachter, naar beneden, positieve hoek,  0.5*pi < hoek < pi
	| a >= (0.5*pi) && b <= zero	= ~b		// rechtsvoor,   naar boven,   positieve hoek,  0      < hoek < 0.5*pi
where
	v	= {dx=base.px-target.px, dy=base.py-target.py}
	d	= dist base target
	a	= acos (max1 (v.dx / d))
	b	= asin (max1 (v.dy / d))
	
	max1 :: Real -> Real
	max1 r
		| r < -1.0	= -1.0
		| r >  1.0	= 1.0
		| otherwise	= r

// gets the angle between two objects
// positive angle is CW, negative is CCW
angleWithObjectForRun :: !(!Position,!Angle) !Position -> Angle 
angleWithObjectForRun (base,angle) target 
	| newAngle >  pi	= newAngle - 2.0*pi
	| newAngle < ~pi	= newAngle + 2.0*pi
	| otherwise			= newAngle
where
	newAngle			= angleWithObject base target - angle
